# -*- coding: utf-8 -*-

"""Task management utilities for executing and tracking multiple tasks."""

from .tasks_manager import TaskResult, TasksManager


__all__ = [
    "TaskResult",
    "TasksManager",
]



# -*- coding: utf-8 -*-

"""Decorators for counting function call invocations."""

from functools import update_wrapper
from typing import Any, Callable


# pylint: disable=too-few-public-methods
class _CountWrapper:
    """It provides the mechanism to count invocations -- class decorator"""

    def __init__(self, fcn: Callable) -> None:
        update_wrapper(self, fcn)
        self.calls_number = 0
        self.fcn = fcn

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.calls_number += 1
        return self.fcn(*args, **kwargs)


def count_calls(fcn: Callable) -> _CountWrapper:
    """
    It provides the mechanism to count invocations...

    :param fcn: The function being decorated.
    :type fcn: Callable
    :return: The wrapped function with calls_number attribute.
    :rtype: _CountWrapper
    """
    return _CountWrapper(fcn)



# -*- coding: utf-8 -*-

"""Interface definitions for factory and task patterns."""

from .factory import IFactory
from .task import ITask, TaskException, TaskStatus


__all__ = [
    "IFactory",
    "ITask",
    "TaskException",
    "TaskStatus",
]



# -*- coding: utf-8 -*-

"""
Singleton pattern decorator for ensuring single
instance classes.
"""

from functools import wraps
from typing import Any, Callable, Optional, Type


# pylint: disable=too-few-public-methods
class _SingletonWrapper:
    """Wrapper class to ensure only one instance of decorated class."""

    def __init__(self, cls: Type) -> None:
        self.cls = cls
        self.instance: Optional[Any] = None
        wraps(cls)(self)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.instance is None:
            self.instance = self.cls(*args, **kwargs)
        return self.instance


def singleton(cls: Type) -> Callable:
    """
    Make a class a Singleton class (only one instance)

    :param cls: The class being decorated.
    :type cls: ``Type``
    :return: The wrapped function.
    :rtype: Callable
    """
    return _SingletonWrapper(cls)



# -*- coding: utf-8 -*-

"""
Compatibility utilities for Python
version differences.
"""

from enum import Enum

try:
    from typing import Self  # pylint: disable=unused-import

except ImportError:
    from typing_extensions import Self  # type: ignore[assignment] # pylint: disable=unused-import


class StrEnum(str, Enum):
    """
    For backward compatibility because StrEnum class
    was added in Python 3.11...
    """

    def __new__(cls, value, *_args, **_kwargs):  # pylint: disable=unused-argument
        if not isinstance(value, str):
            raise TypeError(f"{value} is not a string")

        obj = str.__new__(cls, value)
        obj._value_ = value
        return obj



# -*- coding: utf-8 -*-

"""Caching decorator for memoizing function results."""

from functools import wraps
from typing import Any, Callable, Dict, Tuple


# pylint: disable=too-few-public-methods
class _CacheWrapper:
    """Wrapper class to maintain cache for decorated function."""

    def __init__(self, fcn: Callable) -> None:
        self.fcn = fcn
        self.cache: Dict[Tuple[Any, ...], Any] = {}
        wraps(fcn)(self)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        cache_key = args + tuple(kwargs.items())
        if cache_key not in self.cache:
            self.cache[cache_key] = self.fcn(*args, **kwargs)
        return self.cache[cache_key]


def cache(fcn: Callable) -> Callable:
    """
    It maintains a cache of previous function call results that can be
    used to optimize the performance...

    :param fcn: The function being decorated.
    :type fcn: Callable
    :return: The wrapped function.
    :rtype: Callable
    """
    return _CacheWrapper(fcn)
